//
//  Compat.cpp
//  DDJRBUSBAudio (Kernel Extension)
//
//

#include <IOKit/usb/USB.h>
#include <IOKit/usb/StandardUSB.h>
#include <IOKit/usb/IOUSBHostInterface.h>

#include "AppleUSBAudioCommon.h"
#include "Compat.h"

IOUSBHostDevice* IOUSBCompat::GetDevice(IOUSBHostInterface* hif)
{
    IOUSBHostDevice* device = NULL;
    IOService* deviceCandidate = hif->getProvider();
    FailIf(hif==NULL, Exit);

    while( deviceCandidate != NULL )
    {
        device = OSDynamicCast(IOUSBHostDevice, deviceCandidate);
        if( device != NULL )
        {
            break;
        }
        
        deviceCandidate = deviceCandidate->getProvider();
    }
    FailIf(device==NULL, Exit);
Exit:
    return device;
}

IOUSBHostPipe* IOUSBCompat::FindNextPipe(IOUSBHostInterface* hif,  IOUSBHostPipe* /*pipe*/, IOUSBFindEndpointRequest* find)
{
    IOUSBHostPipe* foundPipe = NULL;
    const EndpointDescriptor* endpointCandidate = NULL;
    FailIf(hif==NULL, Exit);
    FailIf(find==NULL, Exit);

    while( ( endpointCandidate = (EndpointDescriptor*) StandardUSB::getNextAssociatedDescriptorWithType( hif->getConfigurationDescriptor(), hif->getInterfaceDescriptor(), endpointCandidate, kDescriptorTypeEndpoint ) ) != NULL )
    {
        if( endpointCandidate->bDescriptorType == kDescriptorTypeEndpoint && StandardUSB::getEndpointType( endpointCandidate ) == find->type && StandardUSB::getEndpointDirection( endpointCandidate ) == find->direction )
        {
            foundPipe = hif->copyPipe( StandardUSB::getEndpointAddress( endpointCandidate ) );
            break;
        }
    }
    FailIf(foundPipe==NULL, Exit);
Exit:
    return foundPipe;
}

IOReturn IOUSBCompat::Read( IOUSBHostPipe* pipe, IOMemoryDescriptor* buffer, UInt32 noDataTimeout, UInt32 completionTimeout, IOByteCount reqCount, IOUSBHostCompletion *completion)
{
    IOReturn ret = kIOReturnError;

    FailIf(pipe==NULL, Exit);
    FailIf(buffer==NULL, Exit);
    FailIf(completion==NULL, Exit);

    ret = pipe->io( buffer, (uint32_t)reqCount, completion, completionTimeout);
    if(ret)
        debugIOLog("Error: io(Read):0x%x", ret);
    FailIf(ret!=kIOReturnSuccess, Exit);
Exit:
    return ret;
}

IOReturn IOUSBCompat::Read( IOUSBHostPipe* pipe, IOMemoryDescriptor * buffer, UInt64 frameStart, UInt32 numFrames, IOUSBHostIsochronousFrame *pFrames, IOUSBHostIsochronousCompletion *completion, UInt32 /*updateFrequency*/ )
{
    IOReturn ret = kIOReturnError;

    FailIf(pipe==NULL, Exit);
    FailIf(buffer==NULL, Exit);
    FailIf(pFrames==NULL, Exit);
    FailIf(completion==NULL, Exit);

    ret = pipe->io( buffer, pFrames, numFrames, frameStart, completion );
    if(ret)
        debugIOLog("Error: io(Read):0x%x", ret);
    FailIf(ret!=kIOReturnSuccess, Exit);
Exit:
    return ret;
}

IOReturn IOUSBCompat::Write( IOUSBHostPipe* pipe, IOMemoryDescriptor* buffer, UInt64 frameStart, UInt32 numFrames, IOUSBHostIsochronousFrame *pFrames, IOUSBHostIsochronousCompletion *completion, UInt32 /*updateFrequency*/ )
{
    IOReturn ret = kIOReturnError;

    FailIf(pipe==NULL, Exit);
    FailIf(buffer==NULL, Exit);
    FailIf(pFrames==NULL, Exit);
    FailIf(completion==NULL, Exit);
    
    ret = pipe->io( buffer, pFrames, numFrames, frameStart, completion );
    if(ret)
        debugIOLog("Error: io(Write):0x%x", ret);
    FailIf(ret!=kIOReturnSuccess, Exit);
Exit:
    return ret;
}

IOReturn IOUSBCompat::SetPipePolicy( IOUSBHostPipe* pipe, UInt16 maxPacketSize )
{
    IOReturn resultCode = kIOReturnNotReady;
    StandardUSB::EndpointDescriptor     endpointDescriptor;
    
    endpointDescriptor = *( pipe->getEndpointDescriptor() );
    endpointDescriptor.wMaxPacketSize = maxPacketSize;
    
    resultCode = pipe->adjustPipe (&endpointDescriptor, NULL);
    if(resultCode!=kIOReturnSuccess)
        debugIOLog ("Error: adjustPipe(%d):0x%x", endpointDescriptor.wMaxPacketSize, resultCode );
    FailIf(resultCode!=kIOReturnSuccess, Exit);
Exit:
    return resultCode;
}

IOReturn IOUSBCompat::ResetDevice( IOUSBHostInterface* hif )
{
    IOReturn ret = kIOReturnSuccess;

    //this->GetDevice( hif )->ResetDevice();
    debugIOLog("ResetDevice() is now called from user-space");

    return ret;
}

IOReturn IOUSBCompat::GetConfiguration ( IOUSBHostInterface* hif, UInt8* currentConfigValue )
{
    IOReturn ret = kIOReturnSuccess;
    uint8_t configNumber  = 0;
    StandardUSB::DeviceRequest request;
    uint32_t bytesTransferred = 0;

    request.bmRequestType = makeDeviceRequestbmRequestType(kRequestDirectionIn, kRequestTypeStandard, kRequestRecipientDevice);
    request.bRequest      = kDeviceRequestGetConfiguration;
    request.wValue        = 0;
    request.wIndex        = 0;
    request.wLength       = sizeof(configNumber);

    ret = this->GetDevice( hif )->deviceRequest(hif, request, &configNumber, bytesTransferred, kUSBHostStandardRequestCompletionTimeout);
    FailIf(ret != kIOReturnSuccess, Exit);

    *currentConfigValue = configNumber;
    debugIOLog("GetConfiguration currentConfigValue:%d", *currentConfigValue);
    
Exit:
    return ret;
}

IOReturn IOUSBCompat::GetLowLatencyOptionsAndPhysicalMask( IOOptionBits* usbOptions, mach_vm_address_t* physicalMask )
{
    IOReturn ret = kIOReturnSuccess;
    
    *physicalMask = 0xFfffFfffFfffF000ULL;
    *usbOptions = 0;

    return ret;
}

IOReturn IOUSBCompat::GetStringDescriptor( IOUSBHostDevice* device, UInt8 index, char* stringBuffer, size_t stringLength, UInt16 lang /*= StandardUSB::kLanguageIDEnglishUS*/ )
{
    IOReturn ret = kIOReturnError;
    
    memset( stringBuffer, '\0', stringLength );
    const   StringDescriptor* stringDescriptor = device->getStringDescriptor(index);
    
    if( stringDescriptor != NULL && stringDescriptor->bLength > StandardUSB::kDescriptorSize )
    {
        ret = StandardUSB::stringDescriptorToUTF8( stringDescriptor, stringBuffer, stringLength );
    }
    return ret;
}

IOReturn IOUSBCompat::GetStringDescriptor( IOUSBHostInterface* hif, UInt8 index, char* stringBuffer, size_t stringLength, UInt16 lang /*= StandardUSB::kLanguageIDEnglishUS*/ )
{
    return GetStringDescriptor( this->GetDevice( hif ), index, stringBuffer, stringLength, lang );
}

IOReturn IOUSBCompat::DeviceRequest( IOUSBHostInterface* hif, IOUSBDevRequest* request, IOUSBCompletion* completion )
{
    IOReturn    ret = kIOReturnError;

    StandardUSB::DeviceRequest new_request;
    
    new_request.bmRequestType = request->bmRequestType;
    new_request.bRequest = request->bRequest;
    new_request.wValue = request->wValue;
    new_request.wIndex = request->wIndex;
    new_request.wLength = request->wLength;
    
    if( completion == NULL )
    {
        uint32_t bytesTransferred;
        ret = hif->deviceRequest(new_request, request->pData, bytesTransferred);
    }
    else
    {
        IOUSBHostCompletion new_completion;
        
        new_completion.owner = completion->target;
        new_completion.action = completion->action;
        new_completion.parameter = completion->parameter;
        
        ret = hif->deviceRequest(new_request, request->pData, &new_completion);
    }
    return ret;
}

IOReturn IOUSBCompat::DeviceRequest( IOUSBHostInterface* hif, IOUSBDevRequestDesc* request, IOUSBCompletion* completion )
{
    IOUSBDevRequest new_request;
    
    new_request.bmRequestType = request->bmRequestType;
    new_request.bRequest = request->bRequest;
    new_request.wValue = request->wValue;
    new_request.wIndex = request->wIndex;
    new_request.wLength = request->wLength;
    new_request.pData = request->pData;
    new_request.wLenDone = request->wLenDone;

    return this->DeviceRequest( hif, &new_request, completion );
}
